# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
PYSICA - PYthon tools for SImulation and CAlculus
=================================================

Modules in the main package
---------------------------

*constants*
    some physical constant sused in various modules and packages
*parameters*
    some parameters used in various modules and packages
    
Sub-packages in the main package
-------------------------------

*functions*
    some general purpose functions
*fortran*
    modules compiled from Fortran using f2py
*managers*
    modules which can be used to manage the input/output of data from/to ascii files
    and to print physical quantities managing the unit prefixes
*analysis*
    modules which can be helpful to manage distribution functions and data histograms
*plasmapro*
    package of tools for simulatio of plasma discharges.

Documentation is also available in the docstrings.
"""

