# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Utility functions

 This subpackage contains some modules which provide several utility functions.

Modules
-------

*physics*
    functions used to calculate or convert some physical quantities
*optical*
    functions for analysis of the optical properties of thin films (absorption, reflection)
*mathematics*
    functions used to calculate averages, derivatives and other calculations
*pdf*
    collection of probability distribution functions used in the package
*statistics*
    some statistical tools
   
"""
